//
//  blc_network.h
//  network
//
//  Created by Arnaud Blanchard on 30/05/2015.
//
//

/**
 @defgroup blc_network network
 Few functions helping with UDP network.
 @{*/
#ifndef BLC_NETWORK_BASE_H
#define BLC_NETWORK_BASE_H

#include <stdio.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/socket.h>
#include "blc_core.h"
#include <string>

/*By default it is a client*/
#define BLC_SERVER 1
#define BLC_UDP4 2
#define BLC_TCP4 4
#define BLC_UDP4_SERVER (BLC_SERVER | BLC_UDP4)
#define BLC_NET_PROTOCOLE (BLC_UDP4 | BLC_TCP4) // FOr now it is use less but latter it will include all the protocoles

/* To be used with readv/writev, sendmsg/recvmsg. */
#define BLC_SET_IOVEC(iovec, variable)  blc_set_iovec(iovec, &(variable), sizeof(variable), NULL)
#define BLC_SET_IOVEC2(iovec, variable1, variable2)  blc_set_iovec(iovec, &(variable1), sizeof(variable1), &(variable2), sizeof(variable2), NULL)
#define BLC_SET_IOVEC3(iovec, variable1, variable2, variable3)  blc_set_iovec(iovec, &(variable1), sizeof(variable1),&(variable2), sizeof(variable2),&(variable3), sizeof(variable3), NULL)


typedef struct blc_network
#ifdef __cplusplus
:blc_mem{
    /**Init a network connection in **mode** **BLC_UDP4** (udp IPv4) for now.
     **address** can be null (usually for a server). Acceptable value for **address**  is either a valid host name or a numeric host address string consisting of a dotted decimal IPv4
     **port_name** is a port number as a string of a decimal number.
     If buffer_size=0, you are responsible for allocating the memory in .data (use .allocate(size) for example).
     */
    
    blc_network() = default;
    blc_network(char const *address_name, char const *port_name, int mode, ssize_t buffer_size);
	~blc_network();
    
    void publish();
    void send(ssize_t send_size=-1);
    void send_buffer(void const* data, size_t size);
    // Return 1 if data has been read, 0 otherwise (timeout)
    bool recv_buffer_timeout(void *buffer, size_t size, uint64_t us_timeout);
    bool recv_timeout(uint64_t us_timeout);
    void recv_buffer_block(void *buffer, size_t size);
    void recv_block();
    void get_remote_address(char *channel_name, socklen_t name_size, char *port_name, socklen_t port_name_size) const;
    int get_mtu();
    void create_ip4_socket(char const *address_name, char const *port_name, const int socktype);
    void allocate_system_size_buffer();
#else
    { blc_mem mem;
#endif
    struct sockaddr_storage remote_address;
    socklen_t remote_address_length;
    int socket_fd;
    size_t received_size;
}blc_network;
    

    
START_EXTERN_C
///Set a **iovev struct** used by sendmsg/recvmsg, with all the buffers (couples pointer/size) in the list. The list **has** to finish by **NULL**.
int blc_set_iovec(struct iovec *iov, void *data, size_t size, ...);
void blc_set_msghdr(struct msghdr *msghdr, void *data, size_t size, ...);
void blc_network_init(blc_network *network, char const *address_name, char const *port_name, int mode);
END_EXTERN_C

#endif
///@}
